/*-
 * Copyright (c) 1995 by Kenji Rikitake. All rights reserved.
 * Redistribution and usage conditions are mentioned below in
 * the copyright statement of The Regents of the University of California.
 *
 * Kenji Rikitake $Id: mailfeed-rmail.c,v 1.6 1995/10/15 07:04:27 kenji Exp $
 */	
/*-
 * Some parts of this source code have been derived from rmail.c
 * which is distributed with 4.4BSD and Eric Allman's sendmail 8.x.
 * The SCCS ID of the latest rmail.c is 
 * "@(#)rmail.c   8.3 (Berkeley) 5/15/95".
 */
/*-
 * Copyright (c) 1988, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char MAILFEED_Copyright[] =
"Copyright (c) 1995 by Kenji Rikitake. All rights reserved.\n";
static char UCB_copyright[] =
"@(#) Copyright (c) 1988, 1993\n\
	The Regents of the University of California.  All rights reserved.\n";
#endif /* not lint */

/*
 * mailfeed-rmail -- MAILFEED mail server
 *
 * 
 * usage: mailfeed-rmail (with no arguments)
 *
 * This program parses MAILFEED headers on the top of the message header,
 * strips the MAILFEED headers, and passes the processed message to
 * sendmail with the parsed envelope information in MAILFEED headers.
 * This program is called from remote mailbox retriever (e.g., pop-receiver).
 *
 *	X-Mailfeed-Sender: <address of the sender>
 *	X-Mailfeed-Recipient-Number: <number of recipients by decimal>
 *	X-Mailfeed-Recipient-Address: <address of the first recipient>
 *	...
 *	X-Mailfeed-Recipient-Address: <address of the last recipient>
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include <errno.h>
#include <paths.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>

#include "mailfeed.h"

#define PROGRAM_NAME "mailfeed-rmail"

void err __P((int, const char *, ...));
void usage __P((void));

int main(argc, argv)
	int argc;
	char *argv[];
{
	int argn, header_state, i, rec_num, pid, status;
	int pdes[2];
	FILE *fp;
	struct stat sb;
	off_t offset;
	char *p;
	char *args[100], buf[2048], lbuf[2048];

	if (argc > 1) {
		usage();
	}

	/* set sendmail's argument list */
	argn = 0;
	args[argn++] = _PATH_SENDMAIL;	/* sendmail pathname */
	args[argn++] = "-oee";		/* no errors, just status */
	/* 
	 * you might as well specify "-odq" here for reducing load,
	 * victimizing the delivery time 
	 */
	args[argn++] = "-odb";		/* deliver in background */
	args[argn++] = "-oi";		/* ignore '.' on a line by itself */
	args[argn++] = "-oMrMAILFEED";	/* set protocol used */

	/* 
	 * Start parsing the MAILFEED header here.
	 * The MAILFEED header should be formatted as:
	 *
	 *      X-Mailfeed-Sender: <address of the sender>
	 *	X-Mailfeed-Recipient-Number: <number of recipients by decimal>
	 *	X-Mailfeed-Recipient-Address: <address of the first recipient>
	 *	...
	 *	X-Mailfeed-Recipient-Address: <address of the last recipient>
	 *
	 * Note: the above headers should be in the exact sequence shown,
	 *	 and should not be omitted.
	 */

	for (offset = 0, header_state = 0; header_state < 3;) {
		/* Get the line and null-terminate it*/
		if (fgets(lbuf, sizeof(lbuf), stdin) == NULL) {
			exit(EX_DATAERR);
		}
		if ((p = strchr(lbuf, '\n')) == NULL) {
			err(EX_DATAERR, "line too long");
		}
		*p = '\0';
		
		switch (header_state) {
		case 0:		/* first line */ 
			/* 
			 * The first line should tell you 
			 * the envelope sender 
			 */
			if (strncmp(lbuf, HEADER_SENDER, 
				    sizeof(HEADER_SENDER) - 1)) {
				err(EX_DATAERR,
				    "missing %s line: %s", 
				    HEADER_SENDER, lbuf);
			}

			/* copy the sender to sendmail args*/
			/* move to the content part of the header */
			p = lbuf + sizeof(HEADER_SENDER);

			/* find the first non-space character */
			while (isspace(*p)) {
				p++;
			}

			/* Err if empty string */
			if (p == '\0') {
				err(EX_DATAERR,
				    "empty sender line: %s",
				    lbuf);
			}

			/* set name of "from" person */
			(void)snprintf(buf, sizeof(buf), "-f%s", p);
			if ((args[argn++] = strdup(buf)) == NULL) {
				err(EX_TEMPFAIL, NULL);
			}

			header_state = 1;

			break;

		case 1:		/* second line */ 
			/* 
			 * The second line should tell you 
			 * the number of recipients 
			 */
			if (strncmp(lbuf, HEADER_REC_NUM, 
				    sizeof(HEADER_REC_NUM) - 1)) {
				err(EX_DATAERR,
				    "missing %s line: %s", 
				    HEADER_REC_NUM, lbuf);
			}

			/* get the number of recipients */
			/* move to the content part of the header */
			p = lbuf + sizeof(HEADER_REC_NUM);

			/* find the first non-space character */
			while (isspace(*p)) {
				p++;
			}

			/* Err if empty string */
			if (p == '\0') {
				err(EX_DATAERR,
				    "empty recipient-number line: %s",
				    lbuf);
			}

			/* obtain the number of recipients */
			rec_num = atoi(p);
			if (rec_num < 1) {
				err(EX_DATAERR,
				    "invalid value of recipients: %d",
				    rec_num);
			}

			header_state = 2;

			break;

		case 2:		/* getting recipients */
			/*
			 * This line should tell you
			 * the address of a recipient
			 */
			if (strncmp(lbuf, HEADER_REC_ADDR, 
				    sizeof(HEADER_REC_ADDR) - 1)) {
				err(EX_DATAERR,
				    "missing %s line: %s", 
				    HEADER_REC_ADDR, lbuf);
			}

			/* copy each recipient to sendmail args*/
			/* move to the content part of the header */
			p = lbuf + sizeof(HEADER_REC_ADDR);

			/* find the first non-space character */
			while (isspace(*p)) {
				p++;
			}

			/* Err if empty string */
			if (*p == '\0') {
				err(EX_DATAERR,
				    "empty recipient line: %s",
				    lbuf);
			}

			/*
			 * don't copy the address beginning with '-'
			 * as it will be passwd to sendmail and
			 * could be interpreted as flags
			 */
			if (p && *p == '-') {
				err(EX_USAGE,
				    "dash precedes argument: %s", p);
			}
			/*
			 * to prevent confusion of sendmail,
			 * wrap < and > around the addres
			 * (helps to pass addrs like @gw1,@gw2:aa$bb)
			 */
			snprintf(buf, sizeof(buf), "<%s>", p);
			/* copy it to the sendmail args */
			if ((args[argn++] = strdup(buf)) == NULL) {
				err(EX_TEMPFAIL, NULL);
			}
				
			/* decrement the counter and exit the loop if 0 */
			if (--rec_num < 1) {
				header_state = 3;
			}

			break;

		default:
			err(EX_DATAERR, "MAILFEED header corrupted");
			break;

		}
		if (offset != (off_t)-1) {
			offset = (off_t)ftell(stdin);
		}
	}
	args[argn] = (char *)NULL;

	/*
	 * If called with a regular file as standard input, seek to the right
	 * position in the file and just exec sendmail.  Could probably skip 
	 * skip the stat, but it's not unreasonable to believe that a failed
	 * seek will cause future reads to fail.
	 */
	if (!fstat(STDIN_FILENO, &sb) && S_ISREG(sb.st_mode)) {  
		if (lseek(STDIN_FILENO, offset, SEEK_SET) != offset)
			err(EX_TEMPFAIL, "stdin seek");
		execv(_PATH_SENDMAIL, args);
		err(EX_OSERR, "%s", _PATH_SENDMAIL);
	}       

	/* open pipe for local mailer */
	if (pipe(pdes) < 0) {
		err(EX_OSERR, NULL);
	}

	switch (pid = vfork()) {
	case -1:	/* Error */
		err(EX_OSERR, NULL);
	case 0:		/* Child */
		if (pdes[0] != STDIN_FILENO) {
			(void)dup2(pdes[0], STDIN_FILENO);
			(void)close(pdes[0]);
		}
		(void)close(pdes[1]);
		/* invoke sendmail */
		execv(_PATH_SENDMAIL, args);
		_exit(127);
		/* NOTREACHED */
	}

	if ((fp = fdopen(pdes[1], "w")) == NULL) {
		err(EX_OSERR, NULL);
	}
	(void)close(pdes[0]);
	
	/* Copy the file to the pipe until the EOF */
	while (fgets(lbuf, sizeof(lbuf), stdin) != NULL) {
		(void)fprintf(fp, "%s", lbuf);
	}

	if (ferror(stdin)) {
		err(EX_TEMPFAIL, "stdin: %s", strerror(errno));
	}

	if (fclose(fp)) {
		err(EX_OSERR, NULL);
	}

	if ((waitpid(pid, &status, 0)) == -1) {
		err(EX_OSERR, "%s", _PATH_SENDMAIL);
	}

	if (!WIFEXITED(status)) {
		err(EX_OSERR,
		    "%s: did not normally terminate", _PATH_SENDMAIL);
	}

	if (WEXITSTATUS(status)) {
		err(status, "%s: terminated with %d (non-zero) status",
		    _PATH_SENDMAIL, WEXITSTATUS(status));
	}	

	exit(EX_OK);
}

void
usage()
{
	(void)fprintf(stderr, 
		      "usage: %s (no other arguments)\n", PROGRAM_NAME);
	exit(EX_USAGE);
}

#ifdef __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif

void
#ifdef __STDC__
err(int eval, const char *fmt, ...)
#else
err(eval, fmt, va_alist)
	int eval;
	const char *fmt;
	va_dcl
#endif
{
	va_list ap;
#if __STDC__
	va_start(ap, fmt);
#else
	va_start(ap);
#endif
	(void)fprintf(stderr, "%s: ", PROGRAM_NAME);
	(void)vfprintf(stderr, fmt, ap);
	va_end(ap);
	(void)fprintf(stderr, "\n");
	exit(eval);
}
